# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import itertools
import re

from .knot import Knot


def remove_kwargs(s):
    return re.sub("[a-z]*=", "", s)


def make_test(p, q):
    knot = Knot(p, q)
    yield "    def test_{}_{}(self):".format(p, q)
    yield "        knot = Knot({}, {})".format(p, q)
    yield "        self.assertEqual("
    yield "            knot.strings,"
    yield "            ["
    for string in knot.strings:
        yield "                String("
        yield "                    {},".format(string.k)
        yield "                    ["
        for segment in string.segments:
            yield remove_kwargs("                        {},".format(segment))
        yield "                    ],"
        yield "                    ["
        for bridge in string.bridges:
            yield "                        Bridge("
            yield remove_kwargs("                            {},".format(bridge.before))
            yield remove_kwargs("                            {},".format(bridge.after))
            yield remove_kwargs("                            {},".format(bridge.tunnel))
            yield "                        ),"
        yield "                    ],"
        yield "                ),"
    yield "            ]"
    yield "        )"


print """# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import unittest

from .knot import Knot, String, Segment, End, Bridge, Tunnel


class KnotTestCase(unittest.TestCase):"""

tests = []
for p, q in sorted(itertools.product(range(1, 13), repeat=2), key=lambda (p, q): (p + q, p)):
    tests.append("\n".join(make_test(p, q)))
print "\n\n".join(tests)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

"""
.. class:: Colorer

    Objects passed as the ``colorer`` argument of :class:`.TurksHead` must implement this interface.
    More precisely, they must implement **one** of the following methods:

    .. py:method:: compute_color_hsv(knot, k, theta, altitude):

        Return a tuple ``(h, s, v)`` of the color to be applied to the ``k``-th string of the ``knot``,
        at angle ``theta``, at the given ``altitude``.

        ``h`` must be between 0 and 360
        ``s`` and ``v`` must be between 0 and 1.

        :param knot: the :class:`.TurksHead` instance.
        :param k: the index of the current string. Between 0 and ``knot.d``.
        :param theta: the angle on the current string. Between 0 and ``2 * knot.q_prime * math.pi``.
        :param altitude: the altitude of the string at this point. Between -1 and 1.

    .. py:method:: compute_color_rgb(knot, k, theta, altitude):

        Return a tuple ``(r, g, b)`` of the color to be applied to the ``k``-th string of the ``knot``,
        at angle ``theta``, at the given ``altitude``.

        ``r``, ``g`` and ``b`` must be between 0 and 1.

        :param knot: the :class:`.TurksHead` instance.
        :param k: the index of the current string. Between 0 and ``knot.d``.
        :param theta: the angle on the current string. Between 0 and ``2 * knot.q_prime * math.pi``.
        :param altitude: the altitude of the string at this point. Between -1 and 1.
"""

import collections
import math
import unittest

import cairo

from .color import hsv_to_rgb
from .knot import Knot
from .knot.knot import String, Segment, End, Bridge, Tunnel
from ._turkshead import Drawer


class DefaultColorer(object):
    """
    The :class:`Colorer` used when you don't provide one explicitly to :class:`.TurksHead`.

    It uses one color per string (spread on the spectrum) and makes it darker when the string goes down.
    """
    def compute_color_hsv(self, knot, k, theta, altitude):
        # @todo Could we (automatically) add this code to the generated doc?
        h = k * 360. / knot.d
        s = 0.5
        v = 0.5 + altitude / 2
        return h, s, v


class TurksHead(object):
    """
    Turk's head knot objects.

    :param bights: the number of times the string touches the outised of the knot.
    :param leads: the number of times the string turns around the center.
    :param inner: the radius of the empty area inside the knot.
    :param outer: the radius of the knot.
    :param line: the width of the string.
    :param colorer: a :class:`.Colorer` instance. If ``None`` or not provided, a :class:`.DefaultColorer` will be used.
    """
    def __init__(self, bights, leads, inner, outer, line, colorer=None):
        knot = Knot(bights, leads)

        self.__p = knot.p
        self.__q = knot.q
        self.__d = knot.d
        self.__p_prime = knot.p_prime
        self.__q_prime = knot.q_prime
        self.__inner_radius = inner
        self.__outer_radius = outer
        self.__line_width = line

        self.__drawer = Drawer(
            knot=self,
            colorer=colorer or DefaultColorer(),
            strings=knot.strings,
        )


    @property
    def p(self):
        """
        The number of bights.
        """
        return self.__p

    @property
    def q(self):
        """
        The number of leads.
        """
        return self.__q

    @property
    def d(self):
        """
        The greatest common divisor of :attr:`~.TurksHead.p` and :attr:`~.TurksHead.q`.
        """
        return self.__d

    @property
    def p_prime(self):
        """
        :attr:`~.TurksHead.p` divided by :attr:`~.TurksHead.d`.
        """
        return self.__p_prime

    @property
    def q_prime(self):
        """
        :attr:`~.TurksHead.q` divided by :attr:`~.TurksHead.d`.
        """
        return self.__q_prime

    @property
    def inner_radius(self):
        """
        The radius of the empty area inside the knot.
        """
        return self.__inner_radius

    @property
    def outer_radius(self):
        """
        The radius of the knot.
        """
        return self.__outer_radius

    @property
    def line_width(self):
        """
        The width of the string.
        """
        return self.__line_width

    def draw(self, ctx):
        """
        Draw the knot on a :class:`cairo.Context`.
        See `documentation of PyCairo <http://www.cairographics.org/documentation/pycairo/2/reference/context.html>`__.
        """
        self.__drawer.draw(ctx)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import optparse

import cairo

from . import TurksHead

parser = optparse.OptionParser()
parser.add_option("--width", type="int", default=800, help="Width")
parser.add_option("--height", type="int", default=600, help="Height")
parser.add_option("--bights", type="int", default=4, help="Bights")
parser.add_option("--leads", type="int", default=3, help="Leads")
parser.add_option("--radius-variation", type="int", default=200, help="Radius variation")
parser.add_option("--line-width", type="int", default=50, help="Line width")
parser.add_option("--output", type="string", default="turkshead.png", help="Output file")
options, arguments = parser.parse_args()

width = options.width
height = options.height
outerRadius = min(width, height) / 2 - 10
innerRadius = outerRadius - options.radius_variation

img = cairo.ImageSurface(cairo.FORMAT_RGB24, width, height)
ctx = cairo.Context(img)
ctx.set_source_rgb(1, 1, 0xBF / 255.)
ctx.paint()
ctx.translate(width / 2, height / 2)
ctx.scale(1, -1)

TurksHead(options.bights, options.leads, innerRadius, outerRadius, options.line_width).draw(ctx)

img.write_to_png(options.output)



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>
"""
Import:

    >>> import cairo
    >>> from DrawTurksHead import TurksHead

Create a Turk's head:

    >>> knot = TurksHead(bights=7, leads=3, inner=50, outer=200, line=30)

Create a `Cairo <http://www.cairographics.org/pycairo/>`__ context:

    >>> img = cairo.ImageSurface(cairo.FORMAT_RGB24, 440, 440)
    >>> ctx = cairo.Context(img)
    >>> ctx.set_source_rgb(1, 1, 1)
    >>> ctx.paint()
    >>> ctx.translate(220, 220)

Draw the Turk's head on the Cairo context:

    >>> knot.draw(ctx)

Draw circles to verify inner and outer radiuses:

    >>> ctx.set_source_rgb(0, 0, 1)
    >>> ctx.arc(0, 0, 50, 0, 6.30)
    >>> ctx.arc(0, 0, 200, 0, 6.30)
    >>> ctx.stroke()

Save the result to a file:

    >>> img.write_to_png("doc/doctest/1.png")

.. figure:: doctest/1.png
    :align: center

    ``doc/doctest/1.png``

Let's define a utility method:

    >>> def draw_to_png(knot, filename):
    ...   img = cairo.ImageSurface(cairo.FORMAT_RGB24, 440, 440)
    ...   ctx = cairo.Context(img)
    ...   ctx.set_source_rgb(1, 1, 1)
    ...   ctx.paint()
    ...   ctx.translate(220, 220)
    ...   knot.draw(ctx)
    ...   img.write_to_png(filename)

You can choose the color of the drawing by providing a :class:`.Colorer`:

    >>> class MyColorer(object):
    ...   def compute_color_hsv(self, knot, k, theta, altitude):
    ...     h = 180 + k * 360 / knot.d
    ...     s = 1
    ...     v = .5 + altitude / 2
    ...     return h, s, v
    >>> knot = TurksHead(
    ...   bights=7, leads=3,
    ...   inner=50, outer=200, line=30,
    ...   colorer=MyColorer()
    ... )
    >>> draw_to_png(knot, "doc/doctest/2.png")

.. figure:: doctest/2.png
    :align: center

    ``doc/doctest/2.png``

Or:

    >>> from DrawTurksHead import DefaultColorer
    >>> class MyColorer(DefaultColorer):
    ...   def compute_color_hsv(self, knot, k, theta, altitude):
    ...     h, s, v = super(MyColorer, self).compute_color_hsv(knot, k, theta, altitude)
    ...     s = 0.25
    ...     v = v / 2
    ...     return h, s, v
    >>> knot = TurksHead(
    ...   bights=7, leads=3,
    ...   inner=50, outer=200, line=30,
    ...   colorer=MyColorer()
    ... )
    >>> draw_to_png(knot, "doc/doctest/3.png")

.. figure:: doctest/3.png
    :align: center

    ``doc/doctest/3.png``

Or:

    >>> import math
    >>> class MyColorer(object):
    ...   def compute_color_rgb(self, knot, k, theta, altitude):
    ...     assert 0 <= theta < 2 * knot.q_prime * math.pi
    ...     b = theta / (2 * knot.q_prime * math.pi)
    ...     return 0, 0, b
    >>> knot = TurksHead(
    ...   bights=7, leads=3,
    ...   inner=50, outer=200, line=30,
    ...   colorer=MyColorer()
    ... )
    >>> draw_to_png(knot, "doc/doctest/4.png")

.. figure:: doctest/4.png
    :align: center

    ``doc/doctest/4.png``
"""

from .turkshead import TurksHead, DefaultColorer



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import collections
import fractions
import math


End = collections.namedtuple("End", "theta, altitude")
Segment = collections.namedtuple("Segment", "begin, end")
Bridge = collections.namedtuple("Bridge", "before, after, tunnel")
Tunnel = collections.namedtuple("Tunnel", "k, before, after")
String = collections.namedtuple("String", "k, segments, bridges")


class Knot(object):
    def __init__(self, p, q):
        self.p = p
        self.q = q
        self.d = fractions.gcd(p, q)
        self._ks = range(self.d)
        self.p_prime = p // self.d
        self.q_prime = q // self.d
        self.__max_theta = 2 * self.p * self.q_prime
        if self.q == 1:
            self.strings = [String(0, [Segment(End(0, 0), End(2 * self.p, 0))], [])]
        else:
            self.strings = list(self.__make_strings())

    def __make_strings(self):
        intersections = list(self.__find_intersections())
        tidied_intersections = self.__tidy_intersections(intersections)
        ends = list(self.__make_ends(tidied_intersections))
        segments = list(self.__make_segments(ends))
        bridges = list(self.__make_bridges(tidied_intersections, segments))
        for k in self._ks:
            yield String(k, segments[k], bridges[k])

    def __find_intersections(self):
        found = 0
        for m in range(self.d):
            for n in range(m, self.d):
                for theta_1, theta_2 in self.__find_strings_intersections(m, n):
                    assert 0 <= theta_1 < self.__max_theta
                    assert 0 <= theta_2 < self.__max_theta
                    assert m != n or theta_1 < theta_2
                    found += 1
                    yield m, n, theta_1, theta_2
        assert found == self.p * (self.q - 1)

    def __find_strings_intersections(self, m, n):
        if m == n:
            min_a = 1
        else:
            min_a = -self.q_prime + 1
        max_a = self.q_prime
        for a in range(min_a, max_a):
            min_b = int(math.ceil(float(abs(a) * self.p - m - n) / self.q))
            max_b = 2 * self.p_prime - int(math.floor(float(abs(a) * self.p + m + n) / self.q))
            for b in range(min_b, max_b):
                theta_1 = b * self.q - a * self.p + m + n
                theta_2 = b * self.q + a * self.p + m + n
                yield theta_1, theta_2

    def __tidy_intersections(self, intersections):
        tidied_intersections = [{} for k in self._ks]
        for m, n, theta_1, theta_2 in intersections:
            tidied_intersections[m][theta_1] = (n, theta_2)
            tidied_intersections[n][theta_2] = (m, theta_1)
        return tidied_intersections

    def __make_ends(self, tidied_intersections):
        ref_ends = list(self.__make_ref_ends(tidied_intersections))
        yield ref_ends
        ref_ends = {e.theta: e for e in ref_ends}
        for k in range(1, self.d):
            yield list(self.__make_string_ends(ref_ends, tidied_intersections, k))

    def __make_ref_ends(self, tidied_intersections):
        for (i, theta) in enumerate(sorted(tidied_intersections[0].iterkeys())):
            yield End(theta, (-1)**i)

    def __make_string_ends(self, ref_ends, tidied_intersections, k):
        for theta in tidied_intersections[k].iterkeys():
            altitude = ref_ends[(theta - 2 * k) % (self.__max_theta)].altitude
            yield End(theta, altitude)

    def __make_segments(self, ends):
        for string_ends in ends:
            yield list(self.__make_string_segments(string_ends))

    def __make_string_segments(self, string_ends):
        for begin, end in zip(string_ends, string_ends[1:]):
            yield Segment(begin, end)
        yield Segment(
            string_ends[-1],
            End(string_ends[0].theta + self.__max_theta, string_ends[0].altitude)
        )

    def __make_bridges(self, tidied_intersections, segments):
        segs_by_begin = [{s.begin.theta: s for s in string_segments} for string_segments in segments]
        segs_by_end = [{s.end.theta: s for s in string_segments} for string_segments in segments]
        for k, string_segments in enumerate(segments):
            yield list(self.__make_string_bridges(tidied_intersections, segs_by_begin, segs_by_end, k, string_segments))

    def __make_string_bridges(self, tidied_intersections, segs_by_begin, segs_by_end, k, string_segments):
        assert len(string_segments) % 2 == 0
        it = iter(string_segments)  # Inspired by grouper in https://docs.python.org/2/library/itertools.html#recipes
        if string_segments[0].end.altitude != 1:
            after = it.next()
            before = self.__rotate_segment(string_segments[-1])
            yield self.__make_bridge(tidied_intersections, segs_by_begin, segs_by_end, k, before, after)
        for before, after in zip(it, it):
            yield self.__make_bridge(tidied_intersections, segs_by_begin, segs_by_end, k, before, after)

    def __rotate_segment(self, segment):
        return Segment(
            End(segment.begin.theta - self.__max_theta, segment.begin.altitude),
            End(segment.end.theta - self.__max_theta, segment.end.altitude),
        )

    def __make_bridge(self, tidied_intersections, segs_by_begin, segs_by_end, k, before, after):
        assert before.end.theta == after.begin.theta
        assert before.end.altitude == after.begin.altitude == 1
        n, theta_2 = tidied_intersections[k][before.end.theta]
        tunnel = self.__make_tunnel(segs_by_begin, segs_by_end, n, theta_2)
        bridge = Bridge(before, after, tunnel)
        if self.p < 3 or self.q < 3:
            return self.__shorten_bridge(bridge)
        else:
            return bridge

    def __make_tunnel(self, segs_by_begin, segs_by_end, n, theta_2):
        if theta_2 in segs_by_end[n]:
            bef_tunnel = segs_by_end[n][theta_2]
        else:
            bef_tunnel = self.__rotate_segment(segs_by_end[n][theta_2 + self.__max_theta])
        aft_tunnel = segs_by_begin[n][theta_2]
        return Tunnel(n, bef_tunnel, aft_tunnel)

    def __shorten_bridge(self, bridge):
        return Bridge(
            Segment(End((bridge.before.begin.theta + bridge.before.end.theta) / 2, 0), bridge.before.end),
            Segment(bridge.after.begin, End((bridge.after.begin.theta + bridge.after.end.theta) / 2, 0)),
            Tunnel(
                bridge.tunnel.k,
                Segment(End((bridge.tunnel.before.begin.theta + bridge.tunnel.before.end.theta) / 2, 0), bridge.tunnel.before.end),
                Segment(bridge.tunnel.after.begin, End((bridge.tunnel.after.begin.theta + bridge.tunnel.after.end.theta) / 2, 0)),
            ),
        )



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import unittest

from ._turkshead import hsv_to_rgb


class HsvToRgbTestCase(unittest.TestCase):
    def test_red(self):
        self.assertEqual(hsv_to_rgb(0., 1., 1.), (1, 0, 0))

    def test_yellow(self):
        self.assertEqual(hsv_to_rgb(60., 1., 1.), (1, 1, 0))

    def test_yellow_after_360(self):
        self.assertEqual(hsv_to_rgb(420., 1., 1.), (1, 1, 0))

    def test_green(self):
        self.assertEqual(hsv_to_rgb(120., 1., 1.), (0, 1, 0))

    def test_cyan(self):
        self.assertEqual(hsv_to_rgb(180., 1., 1.), (0, 1, 1))

    def test_blue(self):
        self.assertEqual(hsv_to_rgb(240., 1., 1.), (0, 0, 1))

    def test_magenta(self):
        self.assertEqual(hsv_to_rgb(300., 1., 1.), (1, 0, 1))

    def test_blacks(self):
        self.assertEqual(hsv_to_rgb(30., 1., 0.), (0, 0, 0))
        self.assertEqual(hsv_to_rgb(210., 1., 0.), (0, 0, 0))

    def test_whites(self):
        self.assertEqual(hsv_to_rgb(30., 0., 1.), (1, 1, 1))
        self.assertEqual(hsv_to_rgb(210., 0., 1.), (1, 1, 1))

    def test_greys(self):
        self.assertEqual(hsv_to_rgb(30., 0., 0.5), (0.5, 0.5, 0.5))
        self.assertEqual(hsv_to_rgb(210., 0., 0.5), (0.5, 0.5, 0.5))

    def test_grey_ishes(self):
        self.assertEqual(hsv_to_rgb(30., 0.5, 0.5), (0.5, 0.375, 0.25))
        self.assertEqual(hsv_to_rgb(210., 0.5, 0.5), (0.25, 0.375, 0.5))



# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import optparse

import cairo

from . import TurksHead

if __name__ == "__main__":
    parser = optparse.OptionParser()
    parser.add_option("--bights", type="int", default=8, help="Bights")
    parser.add_option("--leads", type="int", default=6, help="Leads")
    parser.add_option("--output", type="string", default="mosaic.png", help="Output file")
    parser.add_option("--base-size", type="int", default=100, help="Increase for bigger image")
    options, arguments = parser.parse_args()

    base_size = options.base_size

    img = cairo.ImageSurface(cairo.FORMAT_RGB24, options.bights * 2 * base_size, options.leads * 2 * base_size)
    ctx = cairo.Context(img)
    ctx.set_source_rgb(1, 1, 1)
    ctx.paint()
    ctx.translate(-base_size, -base_size)

    for p in range(1, options.bights + 1):
        for q in range(1, options.leads + 1):
            ctx.save()
            ctx.translate(p * 2 * base_size, q * 2 * base_size)
            TurksHead(p, q, base_size // 4, 9 * base_size // 10, base_size // 10).draw(ctx)
            ctx.restore()

    img.write_to_png(options.output)



#!/usr/bin/env python
# coding: utf8

# Copyright 2013-2015 Vincent Jacques <vincent@vincent-jacques.net>

import setuptools
import subprocess

version = "0.3.2"


def parse_pkg_config(*args):
    return [
        s[2:]
        for s
        in subprocess.check_output(["pkg-config"] + list(args)).strip().split(" ")
    ]

# @todo Support Python 3 (on Travis as well)

setuptools.setup(
    name="DrawTurksHead",
    version=version,
    description="Draw... Turk's head knots!",
    author="Vincent Jacques",
    author_email="vincent@vincent-jacques.net",
    url="http://pythonhosted.org/DrawTurksHead",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
    ],
    ext_modules=[
        setuptools.Extension(
            "DrawTurksHead._turkshead",
            ["DrawTurksHead/_turkshead.cpp"],
            include_dirs=parse_pkg_config("pycairo", "cairomm-1.0", "--cflags-only-I"),
            libraries=["boost_python"] + parse_pkg_config("cairomm-1.0", "--libs-only-l"),
        ),
    ],
    test_suite="DrawTurksHead.tests",
    command_options={
        "build_sphinx": {
            "version": ("setup.py", version),
            "release": ("setup.py", version),
            "source_dir": ("setup.py", "doc"),
        },
    },
)
